/*
This program creates the following pipelines using a Java process
as the parent (it doesn't matter if the children processes are written
in Java or C or any other language).
                                                    parent
                                                +------------+
                                                |            |
stdin--+---------------------------------------->>0        1>>--->stdout
       |                                        |            |
       |      child_1                 pipe      |            |
       |   +------------+   +------0========0--->>3        2>>--->stderr
       |   |            |   |                   |            |
       +--->>0        1>>---+   +--0========0--->>4          |
       |   |            |       |     pipe      |            |
       |   |          2>>       |               +------------+
       |   |            |       |           CreateTwoPipedChildren.class
       |   +------------+       |
       |   weird_echo.exe       |
       |                        |
       |                        |
       |      child_2           |
       |   +------------+       |
       |   |            |       |
       +--->>0        1>>-------+
           |            |
           |          2>>
           |            |
           +------------+
           weird_echo.exe

To set up these pipelines, we need to create each child process,
have each child inherit its stdin from the parent, and then
connect each child's stdout to a (new) input stream in the
parent (surprisingly in Java there is no need to explicitly
create "pipe" objects).

Notice how the parent's stdin ends up being connected to all three
processes. This is not ideal. In the Windows C version of this program,
each process has its own independent input stream. I cannot figure out
how to do that with Java.

This program (the parent process) reads from its three input streams
(denoted 0, 3, and 4) sequentially. That is, it reads all the input
(up to eof) from stream 3, then it reads all the input (up to eof) from
stream 4, and finally it reads all of the input from stream 0.

The next few versions of this program will read from the three input
streams in parallel.
*/

import java.io.IOException;
import java.util.Scanner;

public class CreateTwoPipedChildren
{
   public static void main(String[] args) throws IOException, InterruptedException
   {
      // Create a command line for running first child.
      ProcessBuilder pb1 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb1 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb1.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process1 = pb1.start();  // this throws IOException


      // Create a command line for running second child.
      ProcessBuilder pb2 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb2 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb2.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process2 = pb2.start();  // this throws IOException


      // Get an input stream that reads from the first child's stdout.
      // This is the first "pipeline" from the first child into the parent.
      Scanner scanner1 = new Scanner( process1.getInputStream() );

      // Get an input stream that reads from the second child's stdout.
      // This is the second "pipeline" from the second child into the parent.
      Scanner scanner2 = new Scanner( process2.getInputStream() );

      // Create a Scanner object to make it easier to use System.in.
      Scanner scanner0 = new Scanner( System.in );


      // Read and filter each line of data from the first child's stdout.
      while ( scanner1.hasNextLine() )
      {
         String oneLine = scanner1.nextLine(); // read from child's stdout

         // Filter the line of input while printing it.
         for (int i = 0; i < oneLine.length(); i++)
         {
            System.out.print( oneLine.charAt(i) );
            if ( -1 != "aeiouAEIOU".indexOf( oneLine.charAt(i) ) )
            {
               System.out.print( oneLine.charAt(i) ); // double vowels
          }
         }
         System.out.println();
      }

      System.out.println("Child 1 ended.");

      // Read and filter each line of data from the second child's stdout.
      while ( scanner2.hasNextLine() )
      {
         String oneLine = scanner2.nextLine(); // read from child's stdout

         // Filter the line of input while printing it.
         for (int i = 0; i < oneLine.length(); i++)
         {
            System.out.print( oneLine.charAt(i) );
            if ( -1 != "aeiouAEIOU".indexOf( oneLine.charAt(i) ) )
            {
               System.out.print( oneLine.charAt(i) ); // triple vowels
               System.out.print( oneLine.charAt(i) );
          }
         }
         System.out.println();
      }

      System.out.println("Child 2 ended.");

      // Read and filter each line of data from the parent's stdin.
      while ( scanner0.hasNextLine() )
      {
         String oneLine = scanner0.nextLine(); // read from parent's stdin

         // Filter the line of input while printing it.
         for (int i = 0; i < oneLine.length(); i++)
         {
            System.out.print( oneLine.charAt(i) );
            if ( -1 != "aeiouAEIOU".indexOf( oneLine.charAt(i) ) )
            {
               System.out.print( oneLine.charAt(i) ); // quadruple vowels
               System.out.print( oneLine.charAt(i) );
               System.out.print( oneLine.charAt(i) );
          }
         }
         System.out.println();
      }

      System.out.println("Parent ended.");

      // Wait for each child to finish its work.
      //process1.waitFor();  // this throws InterruptedException
      //process2.waitFor();  // this throws InterruptedException
   }
}
