/*
This program creates the following pipelines using a Java process
as the parent (it doesn't matter if the children processes are written
in Java or C or any other language).
                                                    parent
                                                +------------+
                                                |            |
stdin--+---------------------------------------->>0        1>>--->stdout
       |                                        |            |
       |      child_1                 pipe      |            |
       |   +------------+   +------0========0--->>3        2>>--->stderr
       |   |            |   |                   |            |
       +--->>0        1>>---+   +--0========0--->>4          |
       |   |            |       |     pipe      |            |
       |   |          2>>       |               +------------+
       |   |            |       |           CreateTwoPipedChildren_Executor.class
       |   +------------+       |
       |   weird_echo.exe       |
       |                        |
       |                        |
       |      child_2           |
       |   +------------+       |
       |   |            |       |
       +--->>0        1>>-------+
           |            |
           |          2>>
           |            |
           +------------+
           weird_echo.exe

To set up these pipelines, we need to create each child process,
have each child inherit its stdin from the parent, and then
connect each child's stdout to a (new) input stream in the
parent (surprisingly in Java there is no need to explicitly
create "pipe" objects).

Notice how the parent's stdin ends up being connected to all three
processes. This is not ideal. In the Windows C version of this program,
each process has its own independent input stream. I cannot figure out
how to do that with Java.

This program (the parent process) reads from its three input streams
(denoted 0, 3, and 4) in parallel, using FutureTask objects and an
Executor to run the tasks. Each FutureTask object that this program
creates has the task of reading and filtering one line from a Scanner
object. The tasks are run by an Executor service with a thread pool
of three threads. So this program does not create more threads than
necessary (unlike CreateTwoPipedChildren_Futures.java).

The FutureTask objects in this program implement Runnable (instead of
Callable). The resulting code is probably a bit more complicated since
each FutureTask does not really return a value.
*/

import java.io.IOException;
import java.util.Scanner;
import java.util.concurrent.FutureTask;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class CreateTwoPipedChildren_Executor_v2
{
   public static void main(String[] args) throws IOException, InterruptedException, ExecutionException
   {
      // Create a command line for running first child.
      ProcessBuilder pb1 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb1 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb1.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process1 = pb1.start();  // this throws IOException


      // Create a command line for running second child.
      ProcessBuilder pb2 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb2 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb2.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process2 = pb2.start();  // this throws IOException


      // Get an input stream that reads from the first child's stdout.
      // This is the first "pipeline" from the first child into the parent.
      Scanner scanner1 = new Scanner( process1.getInputStream() );

      // Get an input stream that reads from the second child's stdout.
      // This is the second "pipeline" from the second child into the parent.
      Scanner scanner2 = new Scanner( process2.getInputStream() );

      // Create a Scanner object to make it easier to use System.in.
      Scanner scanner0 = new Scanner( System.in );


      ExecutorService pool = Executors.newFixedThreadPool(3);

      Task run1 = new Task(scanner1, 2); // double vowels
      Task run2 = new Task(scanner2, 3); // triple vowels
      Task run0 = new Task(scanner0, 4); // quadruple vowels

      // The following three lines come from,
      // https://docs.oracle.com/en/java/javase/11/docs/api/java.base/java/util/concurrent/FutureTask.html#%3Cinit%3E(java.lang.Runnable,V)
      FutureTask<?> task1 = new FutureTask<Void>(run1, null);
      FutureTask<?> task2 = new FutureTask<Void>(run2, null);
      FutureTask<?> task0 = new FutureTask<Void>(run0, null);

      pool.execute(task1); // put the three initial tasks into the executor's work queue
      pool.execute(task2);
      pool.execute(task0);

      boolean child1EOF = false;
      boolean child2EOF = false;
      boolean parentEOF = false;

      // The following loop polls the three FutureTask objects to see
      // when one of them has been completed by the Executor. When a
      // FutureTask object has been completed, a new FutureTask object
      // is created and placed in the Executor's work queue. One of the
      // Executor's threads will pull the task out of the work queue and
      // execute the task (each task reads just one line from its Scanner
      // object). Notice that we are instantiating one FutureTask object
      // for each line of input, which seems to be a lot of objects, but
      // this does not affect the JVM.

      // The Executor's thread pool contains exactly three threads, so
      // all of the work is done by three fixed threads. This is unlike
      // CreateTwoPipedChildren_Futures.java which creates a thread for
      // every line of input, which could be real burden on the operating
      // system (since it is the OS that actually creates threads).

      // This program combines the advantages of
      //     CreateTwoPipedChildren_Threaded.java
      // with the advantages of
      //     CreateTwoPipedChildren_Futures.java
      // We have just three fixed threads (which reduces stress on the OS),
      // and we have the threads working on small tasks (which can improve
      // load balancing on the CPU's).

      // Notice that the main() thread is the thread that does the polling
      // and the instantiating of new FutureTask objects. But polling is a
      // very CPU intensive operation. This is a real waste of CPU time.
      // We need to find a way to avoid polling. The main thread should
      // have a way to block and wait until some task has been completed.
      // In CreateTwoPipedChildren_CompletionService.java we will solve this
      // problem by using the take() method in a ExecutorCompletionService
      // object.
      while( ! parentEOF || ! child1EOF || ! child2EOF )
      {
         if( ! parentEOF && task0.isDone() )
         {
            if (! run0.eof)
            {
               task0 = new FutureTask<Void>(run0, null); // quadruple vowels
               pool.execute(task0);
            }
            else
            {
               parentEOF = true;
               System.out.println("Parent ended.");
            }
            //System.out.println( pool.toString() );
         }
         if( ! child1EOF && task1.isDone() )
         {
            if (! run1.eof)
            {
               task1 = new FutureTask<Void>(run1, null); // double vowels
               pool.execute(task1);
            }
            else
            {
               child1EOF = true;
               System.out.println("Child 1 ended.");
            }
            //System.out.println( pool.toString() );
         }
         if( ! child2EOF && task2.isDone() )
         {
            if (! run2.eof)
            {
               task2 = new FutureTask<Void>(run2, null); // triple vowels
               pool.execute(task2);
            }
            else
            {
               child2EOF = true;
               System.out.println("Child 2 ended.");
            }
            //System.out.println( pool.toString() );
         }
      }
      System.out.println( pool.toString() );
      pool.shutdown();
   }
}



// A Task object, when it is called by one of the threads
// in the Executor's thread pool, will read, filter, and print
// one line of input from the Scanner passed to its constructor.
class Task implements Runnable
{
   private final Scanner scanner;
   private final int count;
   public boolean eof;

   public Task(final Scanner scanner, final int count)
   {
      this.scanner = scanner;
      this.count = count;
      this.eof = false;
   }

   public void run()
   {
      // Read one line of data from the Scanner object.
      if ( scanner.hasNextLine() ) // this blocks
      {
         String oneLine = scanner.nextLine();

         // Filter the line of input while printing it.
         for (int i = 0; i < oneLine.length(); i++)
         {
            System.out.print( oneLine.charAt(i) );
            if ( -1 != "aeiouAEIOU".indexOf( oneLine.charAt(i) ) )
            {
               for (int j = 0; j < count - 1; j++)
               {
                  System.out.print( oneLine.charAt(i) ); // duplicate vowels
               }
            }
         }
         System.out.println();
      }
      else
      {
         eof = true;
      }
   }
}//Task
